/* This file is part of nanolisp, a very small Lisp system.

Copyright (C) 2007  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include "nanolisp.h"

#define NANOLISP_USE_BOEHM_GC
//#define NANOLISP_VERBOSE_GC

#ifdef NANOLISP_USE_BOEHM_GC
#include <gc/gc.h>
#endif // #ifdef NANOLISP_USE_BOEHM_GC

void initialize_nanolisp_garbage_collector(void){
#ifdef NANOLISP_USE_BOEHM_GC
  /* Initialize the Boehm garbage collector, in incremental mode: */
  GC_init();
  GC_enable_incremental();
#endif // #ifdef NANOLISP_USE_BOEHM_GC
}

void* nanolisp_make_garbage_collectable_buffer(const size_t size_in_bytes){
#ifdef NANOLISP_USE_BOEHM_GC
  //printf("%i ", (int)size_in_bytes);
  /* Just make the garbage collector procedure do all the work: */
  void *result = GC_malloc(size_in_bytes);
  return result;
#else
  /* This is only for testing: of course it's not acceptable to make all
     heap-allocated buffers become garbage! */
  void *result = malloc(size_in_bytes);
  assert(result != NULL);
  return result;
#endif // #ifdef NANOLISP_USE_BOEHM_GC
}

#if defined(NANOLISP_USE_BOEHM_GC) && defined(NANOLISP_VERBOSE_GC)
/* This may be helpful when debugging: say which object is about to be
   collected before destroying it: */
static void finalize_sexpression(GC_PTR object, GC_PTR unused_client_data){
  printf("Garbage-collecting ");
  nanolisp_print((nanolisp_sexpression_t)object);
  printf("\n");
}
#endif

nanolisp_sexpression_t nanolisp_make_uninitialized_sexpression(void){
  nanolisp_sexpression_t result =
    nanolisp_make_garbage_collectable_buffer(
      sizeof(nanolisp_sexpression_structure_struct_t));
#if defined(NANOLISP_USE_BOEHM_GC) && defined(NANOLISP_VERBOSE_GC)
  GC_finalization_proc unused_old_finalizer;
  GC_PTR unused_old_data;
  GC_register_finalizer(result, finalize_sexpression, NULL,
                        &unused_old_finalizer, &unused_old_data);
#endif
  return result;
}

void nanolisp_garbage_collect(void){
#ifdef NANOLISP_USE_BOEHM_GC
  printf("Performing a full garbage collection... ");
  fflush(NULL);
  GC_gcollect();
  printf("done.\n");
#else
  printf("===============================================================\n");
  printf("WARNING:\n");
  printf("  NANOLISP_USE_BOEHM_GC was not #define'd at compile time.\n");
  printf("  No garbage collection is being performed.\n");
  printf("===============================================================\n");
#endif // #ifdef NANOLISP_USE_BOEHM_GC
}
