/* This file is part of nanolisp, a very small Lisp system.

Copyright (C) 2007  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include "nanolisp.h"
#include "interpreter.h"

static nanolisp_sexpression_t if_symbol;
static nanolisp_sexpression_t quote_symbol;
static nanolisp_sexpression_t lambda_symbol;
static nanolisp_sexpression_t begin_symbol;
static nanolisp_sexpression_t define_symbol;
static nanolisp_sexpression_t undefine_symbol;

void initialize_interpreter(void){
  if_symbol = nanolisp_make_symbol("if");
  quote_symbol = nanolisp_make_symbol("quote");
  lambda_symbol = nanolisp_make_symbol("lambda");
  begin_symbol = nanolisp_make_symbol("begin");
  define_symbol = nanolisp_make_symbol("define");
  undefine_symbol = nanolisp_make_symbol("undefine");
}

/* Return the value bound to the given key in the given alist, or NULL
   if the key is not in the alist: */
nanolisp_sexpression_t nanolisp_lookup_alist(nanolisp_sexpression_t key,
                                             nanolisp_sexpression_t alist){
  nanolisp_sexpression_t p = alist;
  while(! nanolisp_is_empty_list(p)){
    nanolisp_sexpression_t pair = nanolisp_car(p);
    if(nanolisp_get_boolean(nanolisp_equal(key, nanolisp_car(pair))))
      return nanolisp_cdr(pair);
    p = nanolisp_cdr(p);
  } // while
  return NULL;
}

/* Return the C value true iff key is bound in the given alist: */
bool nanolisp_is_bound_in_alist(nanolisp_sexpression_t key,
                                nanolisp_sexpression_t alist){
  nanolisp_sexpression_t p = alist;
  while(! nanolisp_is_empty_list(p)){
    nanolisp_sexpression_t pair = nanolisp_car(p);
    if(nanolisp_get_boolean(nanolisp_equal(key, nanolisp_car(pair))))
      return true;
    p = nanolisp_cdr(p);
  } // while
  return false;
}

/* Return the length of the given proper list as a C integer: */
int nanolisp_length(nanolisp_sexpression_t list){
  nanolisp_sexpression_t p = list;
  int result = 0;
  while(! nanolisp_is_empty_list(p)){
    result ++;
    p = nanolisp_cdr(p);
  } // while
  return result;
}

/* Return the value bound to the symbol in the given environment; if the
   given symbol is not bound in the given environment then try to look it up
   in the global environment. If no value is found even there then fail: */
nanolisp_sexpression_t nanolisp_lookup(nanolisp_sexpression_t symbol,
                                       nanolisp_sexpression_t environment){
  nanolisp_sexpression_t value_in_environment =
    nanolisp_lookup_alist(symbol, environment);
  if(value_in_environment != NULL)
    return value_in_environment;
  else
    return nanolisp_lookup_global(symbol);
}

/* Evaluate a list of expressions, using the same environment for everyone; return
   the list of results: */
nanolisp_sexpression_t nanolisp_eval_list(nanolisp_sexpression_t expressions,
                                          nanolisp_sexpression_t environment){
  if(nanolisp_is_empty_list(expressions))
    return nanolisp_make_empty_list();
  else
    return nanolisp_make_cons(nanolisp_eval(nanolisp_car(expressions),
                                            environment),
                              nanolisp_eval_list(nanolisp_cdr(expressions),
                                                 environment));
}

nanolisp_sexpression_t nanolisp_eval_combination(nanolisp_sexpression_t operator,
                                                 nanolisp_sexpression_t operands,
                                                 nanolisp_sexpression_t environment){
  /* Here I'm relying on the uniqueness of symbols: for example there is only
     one symbol with the name "if", hence in this case it's safe to use equality
     by identity, which is more efficient: */
  if(operator == if_symbol){
    /* (if CONDITION THEN ELSE) */
    if(nanolisp_length(operands) != 3)
      nanolisp_error("if should have exactly 3 operands: condition, then, else");
    nanolisp_sexpression_t condition = nanolisp_eval(nanolisp_car(operands),
                                                     environment);
    if(nanolisp_get_boolean(condition))
      return nanolisp_eval(nanolisp_cadr(operands), environment);
    else
      return nanolisp_eval(nanolisp_caddr(operands), environment);
  }
  else if(operator == quote_symbol){
    /* (quote X), which is typically written 'X */
    if(nanolisp_length(operands) != 1)
      nanolisp_error("quote should have exactly 1 operand");
    return nanolisp_car(operands); // no evaluation here, of course!
  }
  else if(operator == begin_symbol){
    /* (begin FORM1 ... FORMn) */
    if(nanolisp_is_empty_list(operands))
      nanolisp_error("begin must have at least one form");
    nanolisp_sexpression_t first_operand = nanolisp_car(operands);
    nanolisp_sexpression_t rest_of_operands = nanolisp_cdr(operands);
    if(nanolisp_is_empty_list(rest_of_operands)){
      //printf("begin: tail-eval'ing the last form "); nanolisp_print(first_operand); printf("\n");
      return nanolisp_eval(first_operand, environment);
    }
    else{
      //printf("begin: eval'ing a form "); nanolisp_print(first_operand); printf("\n");
      nanolisp_eval(first_operand, environment);
      //printf("begin: tail-eval'ing the rest "); nanolisp_print(rest_of_operands); printf("\n");
      return nanolisp_eval_combination(operator,
                                       rest_of_operands,
                                       environment);
    } // else
  }
  else if(operator == lambda_symbol){
    /* (lambda ARGUMENTS BODY) */
    if(nanolisp_length(operands) < 2)
      nanolisp_error("lambda should have a list of arguments and a non-empty body");
    nanolisp_sexpression_t formals = nanolisp_car(operands);
    if(! nanolisp_is_list(formals))
      nanolisp_error("parameters should be represented as a non-dotted list of symbols");
    nanolisp_sexpression_t body = nanolisp_make_cons(begin_symbol,
                                                     nanolisp_cdr(operands));
    return nanolisp_make_procedure(formals, body, environment);
  }
  else if(operator == define_symbol){
    /* (define NAME DEFINITION), or (define (NAME FORMAL1 ... FORMALN) FORM1 ... FORMM) */
    if(nanolisp_length(operands) < 2)
      nanolisp_error("define should have at least 2 arguments");
    nanolisp_sexpression_t name = nanolisp_car(operands);
    if(nanolisp_is_symbol(name) && (nanolisp_length(operands) != 2))
      nanolisp_error("define should have exactly 2 arguments when its first argument is a symbol");
    if(nanolisp_is_symbol(name)){
      nanolisp_sexpression_t definition = nanolisp_cadr(operands);
      nanolisp_sexpression_t value = nanolisp_eval(definition, environment);
      nanolisp_set_global(name, value);
      printf("%s was defined as ", nanolisp_get_symbol_as_c_string(name));
      nanolisp_print(value);
      printf("\n");
      return nanolisp_make_boolean(true);
    }
    else{
      /* (define (NAME FORMAL1 .., FORMALN) FORM1 ... FORMM) is just an abbreviation for
         (define NAME (lambda (FORMAL1 ... FORMALN) FORM1 ... FORMM)) */
      nanolisp_sexpression_t name = nanolisp_caar(operands);
      nanolisp_sexpression_t formals = nanolisp_cdar(operands);
      nanolisp_sexpression_t body = nanolisp_cdr(operands);
      nanolisp_sexpression_t lambda_expression =
        nanolisp_make_cons(lambda_symbol,
                           nanolisp_make_cons(formals,
                                              body));
      return nanolisp_eval_combination(
               define_symbol,
               nanolisp_make_list2(name, lambda_expression),
               environment);
    } // end of the case for (define NAME (lambda (FORMAL1 ... FORMALN) FORM1 ... FORMM))
  }
  else if(operator == undefine_symbol){
    if(nanolisp_length(operands) != 1)
      nanolisp_error("undefine should have exactly 1 arguments, the symbol to undefine");
    nanolisp_sexpression_t name = nanolisp_car(operands);
    nanolisp_set_global(name, NULL);
    printf("%s was undefined\n", nanolisp_get_symbol_as_c_string(name));
    return nanolisp_make_boolean(true);
  }
  else{
    /* Procedure application */
    /* Evaluate the operator: */
    nanolisp_sexpression_t procedure = nanolisp_eval(operator, environment);
    /* Evaluate the operands: */
    nanolisp_sexpression_t actuals = nanolisp_eval_list(operands, environment);
    /* Call apply: */
    return nanolisp_apply(procedure, actuals);
  }
}

nanolisp_sexpression_t nanolisp_eval(nanolisp_sexpression_t expression,
                                     nanolisp_sexpression_t environment){
  switch(expression->type){
  /* S-expression of these types are self-evaluating: */
  case nanolisp_integer_type:
  case nanolisp_boolean_type:
  case nanolisp_procedure_type:
  case nanolisp_primitive_procedure_type:
    return expression;
  /* Symbols are evaluated as variables: we should look them up in the given
     environment and, in case of failure, in the global environment: */
  case nanolisp_symbol_type:
    return nanolisp_lookup(expression, environment);
  /* Evaluating the empty list is wrong: */
  case nanolisp_empty_list_type:
    nanolisp_error("the empty list is not a valid combination");
  case nanolisp_cons_type:
    return nanolisp_eval_combination(nanolisp_car(expression),
                                     nanolisp_cdr(expression),
                                     environment);
  default:
    assert(false);
  } // switch
}

/* Both the procedure and the operands must be already evaluated: */
nanolisp_sexpression_t nanolisp_apply(nanolisp_sexpression_t procedure,
                                      nanolisp_sexpression_t parameters){
  switch(procedure->type){
  case nanolisp_primitive_procedure_type:{
    /* Just check the number of parameters, extract the function pointer and
       call it: */
    int number_of_actuals = nanolisp_length(parameters);
    nanolisp_primitive_procedure_structure_t *primitive_procedure_structure =
      nanolisp_get_primitive_procedure(procedure);
    int arity = primitive_procedure_structure->arity;
    if(number_of_actuals < arity)
      nanolisp_error("too few actuals for primitive procedure");
    else if(number_of_actuals > arity)
      nanolisp_error("too many actuals for primitive procedure");
    nanolisp_primitive_procedure_t primitive_procedure =
      primitive_procedure_structure->primitive_procedure;
    switch(arity){
    case 0:
      return primitive_procedure();
    case 1:
      return primitive_procedure(nanolisp_car(parameters));
    case 2:
      return primitive_procedure(nanolisp_car(parameters),
                                 nanolisp_cadr(parameters));
    case 3:
      return primitive_procedure(nanolisp_car(parameters),
                                 nanolisp_cadr(parameters),
                                 nanolisp_caddr(parameters));
    case 4:
      return primitive_procedure(nanolisp_car(parameters),
                                 nanolisp_cadr(parameters),
                                 nanolisp_caddr(parameters),
                                 nanolisp_cadddr(parameters));
    default:
      nanolisp_error("primitive procedures with more than 4 parameters are not currently supported");
    } // switch
  }
  case nanolisp_procedure_type:{
    /* Extract the components we need from the closure: */
    nanolisp_sexpression_t formals = procedure->content.procedure_content.parameters;
    nanolisp_sexpression_t code = procedure->content.procedure_content.code;
    nanolisp_sexpression_t environment = procedure->content.procedure_content.environment;
    /* Make an extended environment which is the same as the environment stored
       in the closure, which one binding added for each parameter: */
    nanolisp_sexpression_t extended_environment = environment;
    nanolisp_sexpression_t remaining_formals = formals;
    nanolisp_sexpression_t remaining_actuals = parameters;
    while(! nanolisp_is_empty_list(remaining_formals)){
      if(nanolisp_is_empty_list(remaining_actuals))
        nanolisp_error("too few actuals for procedure");
      nanolisp_sexpression_t next_formal = nanolisp_car(remaining_formals);
      nanolisp_sexpression_t next_actual = nanolisp_car(remaining_actuals);
      remaining_formals = nanolisp_cdr(remaining_formals);
      remaining_actuals = nanolisp_cdr(remaining_actuals);
      extended_environment = nanolisp_make_cons(nanolisp_make_cons(next_formal,
                                                                   next_actual),
                                                extended_environment);
    } // while
    if(! nanolisp_is_empty_list(remaining_actuals))
      nanolisp_error("too many actuals for procedure");
    //printf("apply: tail-eval'ing on "); nanolisp_print(code); printf("\n");
    //printf("       extended_environment is "); nanolisp_print(extended_environment); printf("\n");
    return nanolisp_eval(code, extended_environment);
  }
  default:
    nanolisp_error("The operator in a combination is not a procedure");
  }
}
