/* This file is part of nanolisp, a very small Lisp system.

Copyright (C) 2007  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include "nanolisp.h"
#include "gc.h"
#include "interpreter.h"

int main(int argc, char **argv){
  if(argc > 2){
    printf("nanolisp should be invoked with either zero or one argument (a file name).\n");
    printf("Exiting.\n");
    exit(EXIT_FAILURE);
  } // if
  initialize_nanolisp_garbage_collector();
  initialize_nanolisp();
  initialize_primitives();
  initialize_interpreter();
  if(argc == 1){
    /* nanolisp was invoked with no parameters, hence the user wants the
       interactive mode: run the REPL: */
    nanolisp_repl();
  }
  else{
    /* The user supplied a file name: let's interpret it and then exit,
       without entering the REPL:*/
    assert(argc == 2);
    /* Open the file and parse it: */
    char *file_name = argv[1];
    FILE *f = fopen(file_name, "r");
    if(f == NULL){
      printf("ERROR: Could not read the file %s. Exiting.\n", file_name);
      exit(EXIT_FAILURE);
    } // if
    
    /* In non-interactive mode when longjmp() is called to manage an error
       we just exit the program: */
    if(setjmp(nanolisp_where_to_jump_in_case_of_error) != 0){
      printf("nanolisp was interrupted by an error in non-interactive mode. Exiting.\n");
      exit(EXIT_FAILURE);
    } // if
    
    nanolisp_sexpression_t sexpressions;
    if(nanolisp_parse_FILE_star(f, &sexpressions) != 0){
      printf("ERROR: parsing the file %s failed. Exiting.\n", file_name);
      exit(EXIT_FAILURE);      
    } // if
    /* Ok, we parsed the file with success, and now we can evaluate the
       s-expressions it contains (note that it contains more than one: 
       sexpressions is a *list* of forms to evaluate): */
    while(! nanolisp_is_empty_list(sexpressions)){
      /* Take the first s-expression and evaluate (note that we don't print
         the result in non-interactive mode): */
      nanolisp_sexpression_t next_sexpression = nanolisp_car(sexpressions);
      nanolisp_eval(next_sexpression, nanolisp_make_empty_list());
      /* Ok, "remove" the first s-expression and loop again: */
      sexpressions = nanolisp_cdr(sexpressions);
    } // while
    
    /* Ok, if we arrived here then everything went smooth, and we can exit
       with success. Note that it's useless to close f just before exiting. */
  } // else
  return 0;
}
