/* This file is part of nanolisp, a very small Lisp system.

Copyright (C) 2007  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include <setjmp.h>
#include <readline/readline.h>
#include "nanolisp.h"
#include "gc.h"

void initialize_nanolisp(void){
  /* Do nothing */
}

static int next_free_index = 0;
/* Implementation of the symbol table: */
struct symbol_table_entry{
  /* The name of this symbol, i.e. the characters representing it: */
  char *string;
  /* The one and only nanolisp object representing this symbol: */
  nanolisp_sexpression_t symbol;
  /* Value bound to the symbol in the global environment, or NULL if the
     symbol is unbound in the global environment: */
  nanolisp_sexpression_t value;
};
struct symbol_table_entry nanolisp_symbol_table[1000]; // To do: make this really dynamic

nanolisp_sexpression_t nanolisp_make_symbol(char *text){
  int i;
  for(i = 0; i < next_free_index; i ++)
    if(!strcmp(text, nanolisp_symbol_table[i].string))
      // the symbol we need is already at the i-th position:
      return nanolisp_symbol_table[i].symbol;
  /* We didn't find it. Ok, let's add the symbol to the table: */
  nanolisp_symbol_table[next_free_index].string = (char*)
    nanolisp_make_garbage_collectable_buffer(strlen(text) + 1);
  strcpy(nanolisp_symbol_table[next_free_index].string, text);
  i = next_free_index;
  next_free_index ++;
  /* Make a new symbol, add it to the table so that the next time
     it's requested we find it, and return it: */
  nanolisp_sexpression_t result =
    nanolisp_make_uninitialized_sexpression();
  result->type = nanolisp_symbol_type;
  result->content.symbol_content = i;
  nanolisp_symbol_table[i].symbol = result;
  nanolisp_symbol_table[i].value = NULL;
  return result;
}

nanolisp_sexpression_t nanolisp_make_integer(int value){
  nanolisp_sexpression_t result =
    nanolisp_make_uninitialized_sexpression();
  result->type = nanolisp_integer_type;
  result->content.integer_content = value;
  return result;
}

nanolisp_sexpression_t nanolisp_make_cons(nanolisp_sexpression_t a,
                                          nanolisp_sexpression_t b){
  nanolisp_sexpression_t result =
    nanolisp_make_uninitialized_sexpression();
  result->type = nanolisp_cons_type;
  result->content.cons_content.car = a;
  result->content.cons_content.cdr = b;
  return result;
}

nanolisp_sexpression_t nanolisp_make_procedure(nanolisp_sexpression_t parameters,
                                               nanolisp_sexpression_t code,
                                               nanolisp_sexpression_t environment){
  nanolisp_sexpression_t result =
    nanolisp_make_uninitialized_sexpression();
  result->type = nanolisp_procedure_type;
  result->content.procedure_content.parameters = parameters;
  result->content.procedure_content.code = code;
  result->content.procedure_content.environment = environment;
  return result;
}

nanolisp_sexpression_t nanolisp_make_primitive_procedure(
                                          int arity,
                                          nanolisp_primitive_procedure_t primitive_procedure){
  nanolisp_sexpression_t result =
    nanolisp_make_uninitialized_sexpression();
  result->type = nanolisp_primitive_procedure_type;
  result->content.primitive_procedure_content.primitive_procedure = primitive_procedure;
  result->content.primitive_procedure_content.arity = arity;
  return result;
}

struct nanolisp_sexpression_structure_struct nanolisp_true = 
  { nanolisp_boolean_type, { true } };

struct nanolisp_sexpression_structure_struct nanolisp_false = 
  { nanolisp_boolean_type, { false } };

struct nanolisp_sexpression_structure_struct nanolisp_empty_list = 
  { nanolisp_empty_list_type, { false } };

nanolisp_sexpression_t nanolisp_make_boolean(bool value){
  if(value)
    return &nanolisp_true;
  else
    return &nanolisp_false;
}

nanolisp_sexpression_t nanolisp_make_empty_list(void){
  return &nanolisp_empty_list;
}

nanolisp_type_t nanolisp_get_type(nanolisp_sexpression_t a){
  return a->type;
}

union nanolisp_expression_content* nanolisp_get_content(nanolisp_sexpression_t a){
  return &(a->content);
}

bool nanolisp_is_integer(nanolisp_sexpression_t a){
  return nanolisp_get_type(a) == nanolisp_integer_type;
}
bool nanolisp_is_symbol(nanolisp_sexpression_t a){
  return nanolisp_get_type(a) == nanolisp_symbol_type;
}
bool nanolisp_is_cons(nanolisp_sexpression_t a){
  return nanolisp_get_type(a) == nanolisp_cons_type;
}
bool nanolisp_is_boolean(nanolisp_sexpression_t a){
  return nanolisp_get_type(a) == nanolisp_boolean_type;
}
bool nanolisp_is_empty_list(nanolisp_sexpression_t a){
  return nanolisp_get_type(a) == nanolisp_empty_list_type;
}
bool nanolisp_is_primitive_procedure(nanolisp_sexpression_t a){
  return nanolisp_get_type(a) == nanolisp_primitive_procedure_type;
}
bool nanolisp_is_procedure(nanolisp_sexpression_t a){
  return nanolisp_get_type(a) == nanolisp_procedure_type;
}

bool nanolisp_is_list(nanolisp_sexpression_t a){
  if(nanolisp_is_empty_list(a))
    return true;
  else if(nanolisp_is_cons(a))
    return nanolisp_is_list(nanolisp_cdr(a));
  else
    return false;
}

void nanolisp_error(const char *error_message){
  printf("ERROR: %s\n", error_message);
  longjmp(nanolisp_where_to_jump_in_case_of_error, 1);
}

/* Check that the given s-expression has the given type, and fail by calling
   nanolisp_error() if this is not the case: */
void nanolisp_assert_type(nanolisp_sexpression_t a, nanolisp_type_t type){
  if(a == NULL){
    printf("nanolisp_assert_type() was called with a NULL pointer.");
    nanolisp_error("a NULL pointer is not a valid s-expression");
  }
  if(nanolisp_get_type(a) != type){
    printf("About the s-expression ");
    nanolisp_print(a);
    printf(":\n");
    nanolisp_error("type error");
  } // if
}

int nanolisp_get_integer(nanolisp_sexpression_t a){
  nanolisp_assert_type(a, nanolisp_integer_type);
  return nanolisp_get_content(a)->integer_content;
}
int nanolisp_get_symbol_index(nanolisp_sexpression_t a){
  nanolisp_assert_type(a, nanolisp_symbol_type);
  return nanolisp_get_content(a)->symbol_content;
}
char* nanolisp_get_symbol_as_c_string(nanolisp_sexpression_t a){
  nanolisp_assert_type(a, nanolisp_symbol_type);
  return nanolisp_symbol_table[nanolisp_get_content(a)->symbol_content].string;
}
nanolisp_cons_t* nanolisp_get_cons(nanolisp_sexpression_t a){
  nanolisp_assert_type(a, nanolisp_cons_type);
  return &(nanolisp_get_content(a)->cons_content);
}
bool nanolisp_get_boolean(nanolisp_sexpression_t a){
  nanolisp_assert_type(a, nanolisp_boolean_type);
  return nanolisp_get_content(a)->boolean_content;
}
nanolisp_primitive_procedure_structure_t* nanolisp_get_primitive_procedure(nanolisp_sexpression_t a){
  nanolisp_assert_type(a, nanolisp_primitive_procedure_type);
  return &(nanolisp_get_content(a)->primitive_procedure_content);
}
nanolisp_procedure_t* nanolisp_get_procedure(nanolisp_sexpression_t a){
  nanolisp_assert_type(a, nanolisp_primitive_procedure_type);
  return &(nanolisp_get_content(a)->procedure_content);
}

nanolisp_sexpression_t nanolisp_read(char *string){
  nanolisp_sexpression_t result;
  if(nanolisp_parse_string(string, &result) != 0)
    nanolisp_error("nanolisp_read() failed (parse error)");
  else if(! nanolisp_is_cons(result))
    nanolisp_error("nanolisp_read() failed (zero s-expressions in the input)");
  else if(! nanolisp_is_empty_list(nanolisp_cdr(result)))
    nanolisp_error("nanolisp_read() failed (more than one s-expression in the input)");
  else
    return nanolisp_car(result);
}

nanolisp_sexpression_t nanolisp_lookup_global(nanolisp_sexpression_t symbol){
  nanolisp_assert_type(symbol, nanolisp_symbol_type);
  nanolisp_sexpression_t result =
    nanolisp_symbol_table[nanolisp_get_content(symbol)->symbol_content].value;
  if(result == NULL){
    printf("About the symbol %s:\n", nanolisp_get_symbol_as_c_string(symbol));
    nanolisp_error("unbound symbol");
  }
  else
    return result;
}

void nanolisp_set_global(nanolisp_sexpression_t symbol,
                         nanolisp_sexpression_t value){
  nanolisp_assert_type(symbol, nanolisp_symbol_type);
  nanolisp_symbol_table[nanolisp_get_content(symbol)->symbol_content].value = value;
}

nanolisp_sexpression_t nanolisp_make_list1(nanolisp_sexpression_t a){
  return nanolisp_make_cons(a, nanolisp_make_empty_list());
}
nanolisp_sexpression_t nanolisp_make_list2(nanolisp_sexpression_t a,
                                           nanolisp_sexpression_t b){
  return nanolisp_make_cons(a,
                            nanolisp_make_cons(b,
                                               nanolisp_make_empty_list()));
}
nanolisp_sexpression_t nanolisp_make_list3(nanolisp_sexpression_t a,
                                           nanolisp_sexpression_t b,
                                           nanolisp_sexpression_t c){
  return nanolisp_make_cons(
           a,
           nanolisp_make_cons(
             b,
             nanolisp_make_cons(c,
                                nanolisp_make_empty_list())));
}
nanolisp_sexpression_t nanolisp_make_list4(nanolisp_sexpression_t a,
                                           nanolisp_sexpression_t b,
                                           nanolisp_sexpression_t c,
                                           nanolisp_sexpression_t d){
  return nanolisp_make_cons(
           a,
           nanolisp_make_cons(
             b,
             nanolisp_make_cons(
               c,
               nanolisp_make_cons(
                 d,
                 nanolisp_make_empty_list()))));
}
