/* This file is part of nanolisp, a very small Lisp system.

Copyright (C) 2007  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include "nanolisp.h"
#include "gc.h"

nanolisp_sexpression_t nanolisp_car(nanolisp_sexpression_t a){
  return nanolisp_get_cons(a)->car;
}
nanolisp_sexpression_t nanolisp_cdr(nanolisp_sexpression_t a){
  return nanolisp_get_cons(a)->cdr;
}

nanolisp_sexpression_t nanolisp_plus(nanolisp_sexpression_t a,
                                     nanolisp_sexpression_t b){
  return nanolisp_make_integer(nanolisp_get_integer(a) +
                               nanolisp_get_integer(b));
}
nanolisp_sexpression_t nanolisp_minus(nanolisp_sexpression_t a,
                                      nanolisp_sexpression_t b){
  return nanolisp_make_integer(nanolisp_get_integer(a) -
                               nanolisp_get_integer(b));
}
nanolisp_sexpression_t nanolisp_times(nanolisp_sexpression_t a,
                                      nanolisp_sexpression_t b){
  return nanolisp_make_integer(nanolisp_get_integer(a) *
                               nanolisp_get_integer(b));
}
nanolisp_sexpression_t nanolisp_divided(nanolisp_sexpression_t a,
                                        nanolisp_sexpression_t b){
  int b_as_int = nanolisp_get_integer(b);
  if(b_as_int == 0)
    nanolisp_error("division by zero");
  return nanolisp_make_integer(nanolisp_get_integer(a) /
                               b_as_int);
}
nanolisp_sexpression_t nanolisp_remainder(nanolisp_sexpression_t a,
                                          nanolisp_sexpression_t b){
  int b_as_int = nanolisp_get_integer(b);
  if(b_as_int == 0)
    nanolisp_error("remainder by zero");
  return nanolisp_make_integer(nanolisp_get_integer(a) %
                               b_as_int);
}

nanolisp_sexpression_t nanolisp_less(nanolisp_sexpression_t a,
                                     nanolisp_sexpression_t b){
  return nanolisp_make_boolean(nanolisp_get_integer(a) <
                               nanolisp_get_integer(b));
}
nanolisp_sexpression_t nanolisp_less_or_equal(nanolisp_sexpression_t a,
                                              nanolisp_sexpression_t b){
  return nanolisp_make_boolean(nanolisp_get_integer(a) <=
                               nanolisp_get_integer(b));
}
nanolisp_sexpression_t nanolisp_greater(nanolisp_sexpression_t a,
                                        nanolisp_sexpression_t b){
  return nanolisp_make_boolean(nanolisp_get_integer(a) >
                               nanolisp_get_integer(b));
}
nanolisp_sexpression_t nanolisp_greater_or_equal(nanolisp_sexpression_t a,
                                                 nanolisp_sexpression_t b){
  return nanolisp_make_boolean(nanolisp_get_integer(a) >=
                               nanolisp_get_integer(b));
}

nanolisp_sexpression_t nanolisp_equal(nanolisp_sexpression_t a,
                                      nanolisp_sexpression_t b){
  bool result;
  /* Particular case: we immediately if a and b are the same object (i.e. if
     their pointers are identical) then of course they are equal: */
  if(a == b)
    result = true;
  /* Other particular case: if a and b have different types then they are
     definitely *not* equal: */
  else if(nanolisp_get_type(a) != nanolisp_get_type(b))
    result = false;
  else
    /* If we arrived here then a and b are distinct objects of the same
       type: */
    switch(nanolisp_get_type(a)){
    case nanolisp_integer_type:
      result = (nanolisp_get_content(a)->integer_content == nanolisp_get_content(b)->integer_content);
      break;
    case nanolisp_symbol_type:
      result = (nanolisp_get_content(a)->symbol_content == nanolisp_get_content(b)->symbol_content);
      break;
    case nanolisp_cons_type:
      result = 
        (nanolisp_equal(nanolisp_get_content(a)->cons_content.car,
                        nanolisp_get_content(b)->cons_content.car) &&
         nanolisp_equal(nanolisp_get_content(a)->cons_content.cdr,
                        nanolisp_get_content(b)->cons_content.cdr));
      break;
    case nanolisp_boolean_type:{
      /* We're comparing C 'booleans', which don't have a specified value for
         'true', but only for 'false'... */
      bool a_content = nanolisp_get_content(a)->boolean_content;
      bool b_content = nanolisp_get_content(b)->boolean_content;
      result = (a_content && b_content) || ((! a_content) && (! b_content));
      break;
    }
    case nanolisp_empty_list_type:
      result = true; // there's no content to compare
      break;
    case nanolisp_primitive_procedure_type:
      nanolisp_error("procedures cannot be compared");
    case nanolisp_procedure_type:
      nanolisp_error("procedures cannot be compared");
    default:
      assert(false);
    } // switch
  return nanolisp_make_boolean(result);
}

nanolisp_sexpression_t nanolisp_display(nanolisp_sexpression_t a){
  nanolisp_print(a);
  return nanolisp_make_boolean(true);
}

nanolisp_sexpression_t nanolisp_newline(void){
  printf("\n");
  return nanolisp_make_boolean(true);
}

nanolisp_sexpression_t nanolisp_read_from_terminal(void){
  static char buffer[10000];
  if(gets(buffer) == NULL)
    nanolisp_error("gets() failed");
  return nanolisp_read(buffer);
}

nanolisp_sexpression_t nanolisp_quit(void){
  exit(EXIT_SUCCESS);
}

nanolisp_sexpression_t nanolisp_integer_p(nanolisp_sexpression_t a){
  return nanolisp_make_boolean(nanolisp_is_integer(a));
}

nanolisp_sexpression_t nanolisp_number_p(nanolisp_sexpression_t a){
  /* As of now the only supported numbers are integers: */
  return nanolisp_make_boolean(nanolisp_is_integer(a));
}

nanolisp_sexpression_t nanolisp_zero_p(nanolisp_sexpression_t a){
  /* As of now the only supported numbers are integers: */
  return nanolisp_make_boolean(nanolisp_is_integer(a) &&
                               (nanolisp_get_integer(a) == 0));
}

nanolisp_sexpression_t nanolisp_symbol_p(nanolisp_sexpression_t a){
  return nanolisp_make_boolean(nanolisp_is_symbol(a));
}

nanolisp_sexpression_t nanolisp_boolean_p(nanolisp_sexpression_t a){
  return nanolisp_make_boolean(nanolisp_is_boolean(a));
}

nanolisp_sexpression_t nanolisp_cons_p(nanolisp_sexpression_t a){
  return nanolisp_make_boolean(nanolisp_is_cons(a));
}

nanolisp_sexpression_t nanolisp_null_p(nanolisp_sexpression_t a){
  return nanolisp_make_boolean(nanolisp_is_empty_list(a));
}

nanolisp_sexpression_t nanolisp_procedure_p(nanolisp_sexpression_t a){
  return nanolisp_make_boolean(nanolisp_is_procedure(a) ||
                               nanolisp_is_primitive_procedure(a));
}

nanolisp_sexpression_t nanolisp_caar(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(a));
}
nanolisp_sexpression_t nanolisp_cadr(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_cdr(a));
}
nanolisp_sexpression_t nanolisp_cdar(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_car(a));
}
nanolisp_sexpression_t nanolisp_cddr(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_cdr(a));
}
nanolisp_sexpression_t nanolisp_caaar(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(nanolisp_car(a)));
}
nanolisp_sexpression_t nanolisp_caadr(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(nanolisp_cdr(a)));
}
nanolisp_sexpression_t nanolisp_cadar(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_cdr(nanolisp_car(a)));
}
nanolisp_sexpression_t nanolisp_caddr(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_cdr(nanolisp_cdr(a)));
}
nanolisp_sexpression_t nanolisp_cdaar(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_car(nanolisp_car(a)));
}
nanolisp_sexpression_t nanolisp_cdadr(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_car(nanolisp_cdr(a)));
}
nanolisp_sexpression_t nanolisp_cddar(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_cdr(nanolisp_car(a)));
}
nanolisp_sexpression_t nanolisp_cdddr(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_cdr(nanolisp_cdr(a)));
}
nanolisp_sexpression_t nanolisp_caaaar(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(nanolisp_car(nanolisp_car(a))));
}
nanolisp_sexpression_t nanolisp_caaadr(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(nanolisp_car(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_caadar(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(nanolisp_cdr(nanolisp_car(a))));
}
nanolisp_sexpression_t nanolisp_caaddr(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(nanolisp_cdr(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_cadaar(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_cdr(nanolisp_car(nanolisp_car(a))));
}
nanolisp_sexpression_t nanolisp_cadadr(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_cdr(nanolisp_car(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_caddar(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_car(nanolisp_cdr(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_cadddr(nanolisp_sexpression_t a){
  return nanolisp_car(nanolisp_cdr(nanolisp_cdr(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_cdaaar(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_car(nanolisp_car(nanolisp_car(a))));
}
nanolisp_sexpression_t nanolisp_cdaadr(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_car(nanolisp_car(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_cdadar(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_car(nanolisp_cdr(nanolisp_car(a))));
}
nanolisp_sexpression_t nanolisp_cdaddr(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_car(nanolisp_cdr(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_cddaar(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_cdr(nanolisp_car(nanolisp_car(a))));
}
nanolisp_sexpression_t nanolisp_cddadr(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_cdr(nanolisp_car(nanolisp_cdr(a))));
}
nanolisp_sexpression_t nanolisp_cdddar(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_cdr(nanolisp_cdr(nanolisp_car(a))));
}
nanolisp_sexpression_t nanolisp_cddddr(nanolisp_sexpression_t a){
  return nanolisp_cdr(nanolisp_cdr(nanolisp_cdr(nanolisp_cdr(a))));
}

nanolisp_sexpression_t nanolisp_gc(void){
  nanolisp_garbage_collect();
  return nanolisp_make_boolean(true);
}

void initialize_primitives(void){
  /* Bind primitive procedures to symbols in the global environment: */
  nanolisp_set_global(nanolisp_make_symbol("cons"),
                      nanolisp_make_primitive_procedure(2, nanolisp_make_cons));
  nanolisp_set_global(nanolisp_make_symbol("car"),
                      nanolisp_make_primitive_procedure(1, nanolisp_car));
  nanolisp_set_global(nanolisp_make_symbol("cdr"),
                      nanolisp_make_primitive_procedure(1, nanolisp_cdr));
  nanolisp_set_global(nanolisp_make_symbol("+"),
                      nanolisp_make_primitive_procedure(2, nanolisp_plus));
  nanolisp_set_global(nanolisp_make_symbol("-"),
                      nanolisp_make_primitive_procedure(2, nanolisp_minus));
  nanolisp_set_global(nanolisp_make_symbol("*"),
                      nanolisp_make_primitive_procedure(2, nanolisp_times));
  nanolisp_set_global(nanolisp_make_symbol("/"),
                      nanolisp_make_primitive_procedure(2, nanolisp_divided));
  nanolisp_set_global(nanolisp_make_symbol("remainder"),
                      nanolisp_make_primitive_procedure(2, nanolisp_remainder));
  nanolisp_set_global(nanolisp_make_symbol("%"),
                      nanolisp_make_primitive_procedure(2, nanolisp_remainder));
  nanolisp_set_global(nanolisp_make_symbol("="),
                      nanolisp_make_primitive_procedure(2, nanolisp_equal));
  nanolisp_set_global(nanolisp_make_symbol("equal?"),
                      nanolisp_make_primitive_procedure(2, nanolisp_equal));
  nanolisp_set_global(nanolisp_make_symbol("<"),
                      nanolisp_make_primitive_procedure(2, nanolisp_less));
  nanolisp_set_global(nanolisp_make_symbol(">"),
                      nanolisp_make_primitive_procedure(2, nanolisp_greater));
  nanolisp_set_global(nanolisp_make_symbol("<="),
                      nanolisp_make_primitive_procedure(2, nanolisp_less_or_equal));
  nanolisp_set_global(nanolisp_make_symbol(">="),
                      nanolisp_make_primitive_procedure(2, nanolisp_greater_or_equal));

  nanolisp_set_global(nanolisp_make_symbol("integer?"),
                      nanolisp_make_primitive_procedure(1, nanolisp_integer_p));
  nanolisp_set_global(nanolisp_make_symbol("number?"),
                      nanolisp_make_primitive_procedure(1, nanolisp_number_p));
  nanolisp_set_global(nanolisp_make_symbol("zero?"),
                      nanolisp_make_primitive_procedure(1, nanolisp_zero_p));
  nanolisp_set_global(nanolisp_make_symbol("symbol?"), nanolisp_make_primitive_procedure(1, nanolisp_symbol_p));
  nanolisp_set_global(nanolisp_make_symbol("boolean?"),
                      nanolisp_make_primitive_procedure(1, nanolisp_boolean_p));
  nanolisp_set_global(nanolisp_make_symbol("cons?"), nanolisp_make_primitive_procedure(1, nanolisp_cons_p));
  nanolisp_set_global(nanolisp_make_symbol("null?"), nanolisp_make_primitive_procedure(1, nanolisp_null_p));
  nanolisp_set_global(nanolisp_make_symbol("empty-list?"), nanolisp_make_primitive_procedure(1, nanolisp_null_p));
  nanolisp_set_global(nanolisp_make_symbol("procedure?"), nanolisp_make_primitive_procedure(1, nanolisp_procedure_p));

  nanolisp_set_global(nanolisp_make_symbol("caar"), nanolisp_make_primitive_procedure(1, nanolisp_caar));
  nanolisp_set_global(nanolisp_make_symbol("cadr"), nanolisp_make_primitive_procedure(1, nanolisp_cadr));
  nanolisp_set_global(nanolisp_make_symbol("cdar"), nanolisp_make_primitive_procedure(1, nanolisp_cdar));
  nanolisp_set_global(nanolisp_make_symbol("cddr"), nanolisp_make_primitive_procedure(1, nanolisp_cddr));
  nanolisp_set_global(nanolisp_make_symbol("caaar"), nanolisp_make_primitive_procedure(1, nanolisp_caaar));
  nanolisp_set_global(nanolisp_make_symbol("caadr"), nanolisp_make_primitive_procedure(1, nanolisp_caadr));
  nanolisp_set_global(nanolisp_make_symbol("cadar"), nanolisp_make_primitive_procedure(1, nanolisp_cadar));
  nanolisp_set_global(nanolisp_make_symbol("caddr"), nanolisp_make_primitive_procedure(1, nanolisp_caddr));
  nanolisp_set_global(nanolisp_make_symbol("cdaar"), nanolisp_make_primitive_procedure(1, nanolisp_cdaar));
  nanolisp_set_global(nanolisp_make_symbol("cdadr"), nanolisp_make_primitive_procedure(1, nanolisp_cdadr));
  nanolisp_set_global(nanolisp_make_symbol("cddar"), nanolisp_make_primitive_procedure(1, nanolisp_cddar));
  nanolisp_set_global(nanolisp_make_symbol("cdddr"), nanolisp_make_primitive_procedure(1, nanolisp_cdddr));
  nanolisp_set_global(nanolisp_make_symbol("caaaar"), nanolisp_make_primitive_procedure(1, nanolisp_caaaar));
  nanolisp_set_global(nanolisp_make_symbol("caaadr"), nanolisp_make_primitive_procedure(1, nanolisp_caaadr));
  nanolisp_set_global(nanolisp_make_symbol("caadar"), nanolisp_make_primitive_procedure(1, nanolisp_caadar));
  nanolisp_set_global(nanolisp_make_symbol("caaddr"), nanolisp_make_primitive_procedure(1, nanolisp_caaddr));
  nanolisp_set_global(nanolisp_make_symbol("cadaar"), nanolisp_make_primitive_procedure(1, nanolisp_cadaar));
  nanolisp_set_global(nanolisp_make_symbol("cadadr"), nanolisp_make_primitive_procedure(1, nanolisp_cadadr));
  nanolisp_set_global(nanolisp_make_symbol("caddar"), nanolisp_make_primitive_procedure(1, nanolisp_caddar));
  nanolisp_set_global(nanolisp_make_symbol("cadddr"), nanolisp_make_primitive_procedure(1, nanolisp_cadddr));
  nanolisp_set_global(nanolisp_make_symbol("cdaaar"), nanolisp_make_primitive_procedure(1, nanolisp_cdaaar));
  nanolisp_set_global(nanolisp_make_symbol("cdaadr"), nanolisp_make_primitive_procedure(1, nanolisp_cdaadr));
  nanolisp_set_global(nanolisp_make_symbol("cdadar"), nanolisp_make_primitive_procedure(1, nanolisp_cdadar));
  nanolisp_set_global(nanolisp_make_symbol("cdaddr"), nanolisp_make_primitive_procedure(1, nanolisp_cdaddr));
  nanolisp_set_global(nanolisp_make_symbol("cddaar"), nanolisp_make_primitive_procedure(1, nanolisp_cddaar));
  nanolisp_set_global(nanolisp_make_symbol("cddadr"), nanolisp_make_primitive_procedure(1, nanolisp_cddadr));
  nanolisp_set_global(nanolisp_make_symbol("cdddar"), nanolisp_make_primitive_procedure(1, nanolisp_cdddar));
  nanolisp_set_global(nanolisp_make_symbol("cddddr"), nanolisp_make_primitive_procedure(1, nanolisp_cddddr));

  nanolisp_set_global(nanolisp_make_symbol("display"),
                      nanolisp_make_primitive_procedure(1, nanolisp_display));
  nanolisp_set_global(nanolisp_make_symbol("newline"),
                      nanolisp_make_primitive_procedure(0, nanolisp_newline));
  nanolisp_set_global(nanolisp_make_symbol("read"),
                      nanolisp_make_primitive_procedure(0, nanolisp_read_from_terminal));

  nanolisp_set_global(nanolisp_make_symbol("quit"),
                      nanolisp_make_primitive_procedure(0, nanolisp_quit));

  nanolisp_set_global(nanolisp_make_symbol("no-warranty"),
                      nanolisp_make_primitive_procedure(0, nanolisp_print_no_warranty));
  nanolisp_set_global(nanolisp_make_symbol("license"),
                      nanolisp_make_primitive_procedure(0, nanolisp_print_license));
  nanolisp_set_global(nanolisp_make_symbol("copying"),
                      nanolisp_make_primitive_procedure(0, nanolisp_print_license));
  nanolisp_set_global(nanolisp_make_symbol("gpl"),
                      nanolisp_make_primitive_procedure(0, nanolisp_print_license));

  nanolisp_set_global(nanolisp_make_symbol("gc"),
                      nanolisp_make_primitive_procedure(0, nanolisp_gc));
}
